/*******************/
/* System includes */
/*******************/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>

/******************/
/* EPICS includes */
/******************/

/* EPICS includes */
#include <drvIpac.h>
#include <iocsh.h>
#include <epicsExport.h>
#include <epicsThread.h>
#include <epicsString.h>
#include <epicsTimer.h>
#include <epicsExit.h>
#include <epicsMutex.h>
#include <epicsMessageQueue.h>
#include <cantProceed.h>
#include <devLib.h>
#include <errlog.h>
#include <asynPortDriver.h>

#ifndef DRVHY8413IPASYN_H
#define DRVHY8413IPASYN_H
#endif

/**************************************************************/
/* A couple of key parameters defining the ADC memory layout. */
#define BANKSIZE  				16384         	/* maximum channel size: 128K */
#define MAX_MESSAGES  			100           	/* maximum number of messages */
#define MAX_ADC_CHANNELS    	16      		/* number of ADC channels on 8413 card */
#define MAX_CARDS          		256          /* maxiumum number of cards in a system */

/* Identifiers in PROM. */
#define VITA4       			"VITA 4"
#define HYTECID_LO  			0x0300
#define HYTECID_HI  			0x0080
#define PROM_MODEL  			0x8413

/* define bits in the 8413 CSR */
#define CSR_F 					0x0001          /* STAUTS: indicate internal FIFO full */
#define CSR_TF	 				0x0002          /* STAUTS: indicate external FIFO full */
#define CSR_FE	 				0x0004          /* STAUTS: indicate internal FIFO empty */
#define CSR_THF 				0x0008          /* STAUTS: indicate external FIFO half full */
#define CSR_RSTFIFO 			0x0010          /* Reset FIFO */
#define CSR_DRE 				0x0020          /* DMA request allowed if set */
#define CSR_TE 					0x0040          /* STAUTS: indicate external FIFO empty */
#define CSR_EHF    				0x0080          /* Enable interrupt on external FIFO half full */
#define CSR_EF     				0x0100          /* Enable interrupt on internal FIFO full */
#define CSR_ETF     			0x0200          /* Enable interrupt on external FIFO full */
#define CSR_EG      			0x0800          /* Enable hardware ARMing via Strobe line input */
#define CSR_XC      			0x1000          /* 0 => use internal clock, 1 => external clock. */
#define CSR_ST      			0x2000          /* Software trigger. To trigger, do "clear" then "set" */
#define CSR_ET      			0x4000          /* 0 => continuous capture, 1 => on trigger. */
#define CSR_ARM     			0x8000          /* Software ARMing. */

/* define extension csr */
#define ACR_RGE 				0x0001          /* Range setting: 0 = +/-10V, 1 = +/-5V */
#define ACR_NS	 				0x0004          /* 0 = standby mode, 1 = normal mode */
#define ACR_2C	 				0x0008          /* 0 = 2's complement, 1 = normal data 0000 ~ FFFF */
#define ACR_10VCAL0_5			0x0010          /* calibration values for +/-10V for channel 0-5 */
#define ACR_10VCAL6_11			0x0020          /* calibration values for +/-10V for channel 6-11 */
#define ACR_10VCAL12_16			0x0030          /* calibration values for +/-10V for channel 12-16 */
#define ACR_5VCAL0_5			0x0040          /* calibration values for +/-5V for channel 0-5 */
#define ACR_5VCAL6_11			0x0050          /* calibration values for +/-5V for channel 6-11 */
#define ACR_5VCAL12_16			0x0060          /* calibration values for +/-5V for channel 12-16 */

#define ENABLEADC 				1
#define DISABLEADC 				0

#define INIT 					0
#define RUNTIME 				1

/* driver support flags */
/* Bit0 - continuos mode support */
/* Bit1 - trigger mode support */
#define DRIVER_SUPPORT     		0x0003   		/* support continuous and trigger modes */

#define WAITING_TRIGGER         0           	/* waiting for trigger */
#define TRIGGERRED_COLLECTING   1           	/* triggerred and collecting data */
#define FINISHED_AVAILABLE      2           	/* finished and data available */

#ifndef USHORT 
#define USHORT unsigned short
#endif
#ifndef ULONG
#define ULONG unsigned int
#endif

/*****************************************************************************/
/*                                                                           */
/*                              Data Structures                              */
/*                                                                           */
/*****************************************************************************/


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* ADC Register and P#define hy8413SupportString			"SUPPORT"
ROM Layout                                              */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The I/O memory is divided into two areas: register and PROM. We separate
 * them out for coding convenience. 
 */

/* This data structure maps the I/O registers */
typedef struct
{
    USHORT csr;                  	/* 0x00 */
    USHORT notused1;                /* 0x02 */
    USHORT notused2;          		/* 0x04 */
    USHORT clockRate;               /* 0x06 */
    USHORT vector;                  /* 0x08 */
    USHORT intFIFO;           		/* 0x0A */
    USHORT extFIFO;               	/* 0x0C */
    USHORT extCounter;             	/* 0x0E */
    USHORT data[MAX_ADC_CHANNELS];  /* 0x10 - 0x2E */
    USHORT ref[2];  				/* 0x30 - 0x32 */
    USHORT acr;               		/* 0x34 */
} REGISTERS;


/* This data structure maps the image of the ID PROM.  The PROM
 * starts at offset 0x80 from the base of the register memory. */
typedef struct
{
    char vita4[6];                  /* 0x80 */
    USHORT hytecIdHigh;             /* 0x86 */
    USHORT hytecIdLow;              /* 0x88 */
    USHORT model;                   /* 0x8A */
    USHORT revision;                /* 0x8C */
    USHORT unused1;                 /* 0x8E */
    USHORT driverIdHigh;            /* 0x90 */
    USHORT driverIdLow;             /* 0x92 */
    USHORT flags;                   /* 0x94 */
    USHORT byteCount;               /* 0x96 */
    USHORT calType;                 /* 0x98 */
    USHORT serialNumber;            /* 0x9A */
    USHORT unused2;                 /* 0x9C */
    USHORT WLO;                     /* 0x9E */
    USHORT CAL[0xBC - 0x9E];		/* store calibration data */
} ID_PROM;

/* Now put them together, the I/O memory map. */
typedef struct
{
    REGISTERS reg;                              /* 0x00 - 0x2F */
    char unused[0x80-sizeof(REGISTERS)];        /* 0x30 - 0x80 */
    ID_PROM prom;                               /* 0x80 - 0x9F */
} IO_MEMORY;

/* enumerated type of the driver's operational modes */
typedef enum{		
	hy8413Continuous,
	hy8413Trigger,
} hy8413Modes;

typedef struct 
{
    USHORT ngFS;                        /* ADC output at -10V (NeGative Full Scale) */
    USHORT ngHS;                        /* ADC output at  -5V (NeGative Half Scale) */
    USHORT zero;                        /* ADC output at   0V (zero) */
    USHORT psHS;                        /* ADC output at  +5V (PoSitive Half Scale) */
    USHORT psFS;                        /* ADC output at +10V (PoSitive Full Scale) */
} CALDATA;

/*****************************************************************************/
/*                                                                           */
/*                      Interrupt/Queue Routines	                         */
/*                                                                           */
/*****************************************************************************/

extern "C" {
static void rebootCallbackC     ( void *drvPvt );
static void INTSRC              ( int cardNumber ); 	/* Interrupt Service Routine */
static void intQueuedTaskC      ( void *drvPvt );
}

/** This is the class definition for the Hy8413Asyn class
  */
class Hy8413Asyn : public asynPortDriver {
public:
	Hy8413Asyn(char *portName, int carrierNum, int ipSlotNum, int vectorNum, int mode, 
			int range, int format, int clockRate, int extClock);

    /* These are the methods that we override from asynPortDriver */
    virtual asynStatus writeInt32(asynUser *pasynUser, epicsInt32 value);
    virtual asynStatus readInt32(asynUser *pasynUser, epicsInt32 *value);
    virtual asynStatus getBounds(asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high);
    virtual asynStatus readFloat64(asynUser *pasynUser, epicsFloat64 *value);
    virtual void report(FILE *fp, int details);
    
    void rebootCallback();
	void intSR(); 				/* Interrupt Service Routine */
	void intQueuedTask();

protected:
    int hy8413Data_;
    #define FIRST_8413_PARAM hy8413Data_
	int hy8413FloatData_;
	int hy8413SetMode_;
	int hy8413Enable_;
	int hy8413SoftTrigger_;
	int hy8413AutoReenable_;
	int hy8413SetClockRate_;
	int hy8413SetExtClock_;
	int hy8413Range_;
	int hy8413SetFormat_;
	int hy8413resetFIFO_;

	int hy8413Counter_;
	int hy8413TriggerStatus_;
	int hy8413FWVersion_;
	int hy8413DriverVersion_;
	int hy8413Support_;
    #define LAST_8413_PARAM hy8413Support_

private:
	void setupADC               ();
	asynStatus setClockRate     ( int clockrate );
	asynStatus setExtClock      ( int extclock );
	asynStatus softwareTrigger  ();
	asynStatus setMode          ( int mode );
	asynStatus setRange         ( int range );
	asynStatus setFormat        ( int format );
	asynStatus resetFIFO        ();
	asynStatus setEnable		( int enable );
	asynStatus setAutoReEnable	( int enable );
	double toVoltage            ( USHORT rawval );
	USHORT calibration			( USHORT rawval, int channel );
    
	volatile IO_MEMORY 	*adc;			   	/* pointer to IP card A16 space registers, ID and calibration data */ 

    ULONG 	rawData[MAX_ADC_CHANNELS][BANKSIZE];/* ADC data for I/O Intr Int32Array */
    double 	adcData[MAX_ADC_CHANNELS][BANKSIZE];/* ADC data for I/O Intr Float64Array */

    CALDATA cal10V[MAX_ADC_CHANNELS];
    CALDATA cal5V[MAX_ADC_CHANNELS];
	/* Storage for the current Configuration parameters */

    USHORT	carrierNum;		                /* carrier index number when adding carrier */
    USHORT	ipSlotNum;		                /* IP card slot number on IP carrier - boot time configuration only */
    USHORT	vectorNum;		                /* Interrupt vector number - boot time configuration only */
    USHORT	mode;		                	/* mode, 0 for continuous mode, 1 for trigger mode */
    USHORT	range;		                	/* range setting, 0 for -10v ~ +10V, 1 for -5V ~ +5V */
    USHORT	format;		                	/* data format, 0 for 2's complement, 1 for straight */
    USHORT	clockrate;	                	/* clock rate */
    USHORT	extclock;		               	/* 0 internal clock, 1 for external clock */

/* driver specific parameters */
    int		rebooting;		                /* Flag to indicate is the IOC is currently re-booting */

/* Work Queue Stuff */
    epicsMessageQueueId intMsgQId;
    int messagesSent;                       /* for report */
    int messagesFailed;                     /* for report */
};

#define hy8413DataString	        "DATA"
#define hy8413FloatDataString	    "FLOATDATA"
#define hy8413RangeString	    	"RANGE"
#define hy8413SetFormatString	    "FORMAT"
#define hy8413ResetFIFOString	   	"RESETFIFO"
#define hy8413SetModeString		    "SETMODE"
#define hy8413EnableString		    "ENABLE"
#define hy8413SoftTriggerString		"SOFTTRIGGER"
#define hy8413AutoReenableString	"REENABLE"
#define hy8413SetClockRateString	"SETCLOCKRATE"
#define hy8413SetExtClockString		"SETEXTCLOCK"

#define hy8413CounterString			"COUNTER"
#define hy8413TriggerStatusString	"TRIGGERSTATE"
#define hy8413FWVersionString		"FWVERSION"
#define hy8413DriverVersionString	"DRIVERVERSION"
#define hy8413SupportString			"SUPPORT"

#define NUM_PARAMS (&LAST_8413_PARAM - &FIRST_8413_PARAM + 1)

/** Configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *    vectorNum = interrupt vector (0 = find one for me),
 *         mode = 0 = continuous, 1 = triggered, 3 = gated,
 *      range   = 0 = -10V ~ +10V, 1 = -5V ~ +5V 
 *		format  = 0 = 2's complement, 1 = straight binary
 *    clockRate = 0  = 1Hz,    1 = 2Hz,    2 = 5Hz,    3 = 10Hz,
 *				  4  = 20Hz,   5 = 50Hz,   6 = 100Hz,  7 = 200Hz,
 *				  8  = 500Hz,  9 = 1kHz,  10 = 2kHz,  11 = 5kHz,
 *				  12 = 10khz, 13 = 20Khz, 14 = 50kHz, 15 = 100kHz.
 *				  16 = 160kHz
 *     extClock = 0 = internal, 1 = extrenal,
 */
extern "C" int Hy8413AsynInit(char *portName, int vmeSlotNum, int ipSlotNum, int vectorNum, 
					int mode, int range, int format, int clockRate, int extClock)
{
    Hy8413Asyn *pHy8413Asyn = new Hy8413Asyn(portName, vmeSlotNum, ipSlotNum, vectorNum, mode, 
    		range, format, clockRate, extClock);
    pHy8413Asyn = NULL;  /* This is just to avoid compiler warnings */
    return(asynSuccess);
}

				
// Init
static const iocshArg initArg0 = { "Port name",iocshArgString};
static const iocshArg initArg1 = { "vmeSlotNum",iocshArgInt};
static const iocshArg initArg2 = { "ipSlotNum",iocshArgInt};
static const iocshArg initArg3 = { "vectorNum",iocshArgInt};
static const iocshArg initArg4 = { "Scan mode", iocshArgInt};
static const iocshArg initArg5 = { "range", iocshArgInt};
static const iocshArg initArg6 = { "format", iocshArgInt};
static const iocshArg initArg7 = { "clockRate", iocshArgInt};
static const iocshArg initArg8 = { "extClock", iocshArgInt};
static const iocshArg * const initArgs[] = {&initArg0,
                                             &initArg1,
                                             &initArg2,
                                             &initArg3,
                                             &initArg4,
                                             &initArg5,
                                             &initArg6,
                                             &initArg7,
                                             &initArg8};
static const iocshFuncDef initFuncDef = {"Hy8413AsynInit",9,initArgs};
static void initCallFunc(const iocshArgBuf *args)
{
    Hy8413AsynInit(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival, args[6].ival, 
					args[7].ival, args[8].ival);
}

void drvHy8413AsynRegister(void)
{
    iocshRegister(&initFuncDef,initCallFunc);
}

extern "C" {
epicsExportRegistrar(drvHy8413AsynRegister);
}

